# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import inspect
import sys
import typing
from unittest import mock

import hikari
import pytest

import tanjun


class TestClient:
    def test_with_listener_no_provided_event(self):
        async def callback(foo) -> None:  # type: ignore
            ...

        add_listener_ = mock.Mock()

        class StubClient(tanjun.Client):
            add_listener = add_listener_

        client = StubClient(mock.Mock())

        with pytest.raises(ValueError, match="Missing event argument annotation"):
            client.with_listener()(callback)  # pyright: ignore[reportUnknownArgumentType]

        add_listener_.assert_not_called()

    def test_with_listener_no_provided_event_callback_has_no_signature(self):
        with pytest.raises(ValueError, match=".+"):
            inspect.Signature.from_callable(int)

        add_listener_ = mock.Mock()

        class StubClient(tanjun.Client):
            add_listener = add_listener_

        client = StubClient(mock.Mock())

        with pytest.raises(ValueError, match="Missing event type"):
            client.with_listener()(int)  # type: ignore

        add_listener_.assert_not_called()

    def test_with_listener_with_type_hint(self):
        async def callback(event: hikari.BanCreateEvent) -> None:
            ...

        add_listener_ = mock.Mock()

        class StubClient(tanjun.Client):
            add_listener = add_listener_

        client = StubClient(mock.Mock())

        result = client.with_listener()(callback)

        assert result is callback
        add_listener_.assert_called_once_with(hikari.BanCreateEvent, callback)

    def test_with_listener_with_type_hint_in_annotated(self):
        async def callback(event: typing.Annotated[hikari.BanCreateEvent, 123, 321]) -> None:
            ...

        add_listener_ = mock.Mock()

        class StubClient(tanjun.Client):
            add_listener = add_listener_

        client = StubClient(mock.Mock())

        result = client.with_listener()(callback)

        assert result is callback
        add_listener_.assert_called_once_with(hikari.BanCreateEvent, callback)

    def test_with_listener_with_positional_only_type_hint(self):
        async def callback(event: hikari.BanDeleteEvent, /) -> None:
            ...

        add_listener_ = mock.Mock()

        class StubClient(tanjun.Client):
            add_listener = add_listener_

        client = StubClient(mock.Mock())

        result = client.with_listener()(callback)

        assert result is callback
        add_listener_.assert_called_once_with(hikari.BanDeleteEvent, callback)

    def test_with_listener_with_var_positional_type_hint(self):
        async def callback(*event: hikari.BanEvent) -> None:
            ...

        add_listener_ = mock.Mock()

        class StubClient(tanjun.Client):
            add_listener = add_listener_

        client = StubClient(mock.Mock())

        result = client.with_listener()(callback)

        assert result is callback
        add_listener_.assert_called_once_with(hikari.BanEvent, callback)

    def test_with_listener_with_type_hint_union(self):
        async def callback(event: typing.Union[hikari.RoleEvent, typing.Literal["ok"], hikari.GuildEvent, str]) -> None:
            ...

        add_listener_ = mock.Mock()

        class StubClient(tanjun.Client):
            add_listener = add_listener_

        client = StubClient(mock.Mock())

        result = client.with_listener()(callback)

        assert result is callback
        add_listener_.assert_has_calls([mock.call(hikari.RoleEvent, callback), mock.call(hikari.GuildEvent, callback)])

    def test_with_listener_with_type_hint_union_nested_annotated(self):
        async def callback(
            event: typing.Annotated[
                typing.Union[
                    typing.Annotated[typing.Union[hikari.RoleEvent, hikari.ReactionDeleteEvent], 123, 321],
                    hikari.GuildEvent,
                ],
                True,
                "meow",
            ]
        ) -> None:
            ...

        add_listener_ = mock.Mock()

        class StubClient(tanjun.Client):
            add_listener = add_listener_

        client = StubClient(mock.Mock())

        result = client.with_listener()(callback)

        assert result is callback
        add_listener_.assert_has_calls(
            [
                mock.call(hikari.RoleEvent, callback),
                mock.call(hikari.ReactionDeleteEvent, callback),
                mock.call(hikari.GuildEvent, callback),
            ]
        )

    # These tests covers syntax which was introduced in 3.10
    if sys.version_info >= (3, 10):

        def test_with_listener_with_type_hint_310_union(self):
            async def callback(event: hikari.ShardEvent | typing.Literal[""] | hikari.VoiceEvent | str) -> None:
                ...

            add_listener_ = mock.Mock()

            class StubClient(tanjun.Client):
                add_listener = add_listener_

            client = StubClient(mock.Mock())

            result = client.with_listener()(callback)

            assert result is callback
            add_listener_.assert_has_calls(
                [mock.call(hikari.ShardEvent, callback), mock.call(hikari.VoiceEvent, callback)]
            )

        def test_with_listener_with_type_hint_310_union_nested_annotated(self):
            async def callback(
                event: typing.Annotated[
                    typing.Annotated[hikari.BanEvent | hikari.GuildEvent, 123, 321] | hikari.InviteEvent, True, "meow"
                ]
            ) -> None:
                ...

            add_listener_ = mock.Mock()

            class StubClient(tanjun.Client):
                add_listener = add_listener_

            client = StubClient(mock.Mock())

            result = client.with_listener()(callback)

            assert result is callback
            add_listener_.assert_has_calls(
                [
                    mock.call(hikari.BanEvent, callback),
                    mock.call(hikari.GuildEvent, callback),
                    mock.call(hikari.InviteEvent, callback),
                ]
            )



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Menu context implementation."""
from __future__ import annotations

__all__: list[str] = ["MenuContext"]

import typing

import hikari

from .. import _internal
from .. import abc as tanjun
from . import slash

if typing.TYPE_CHECKING:
    import asyncio
    from collections import abc as collections

    from typing_extensions import Self

    _T = typing.TypeVar("_T")
    _ResponseTypeT = typing.Union[
        hikari.api.InteractionMessageBuilder, hikari.api.InteractionDeferredBuilder, hikari.api.InteractionModalBuilder
    ]


_VALID_TYPES: frozenset[typing.Literal[hikari.CommandType.USER, hikari.CommandType.MESSAGE]] = frozenset(
    [hikari.CommandType.USER, hikari.CommandType.MESSAGE]
)


class MenuContext(slash.AppCommandContext, tanjun.MenuContext):
    """Standard menu command execution context."""

    __slots__ = ("_command", "_marked_not_found", "_on_not_found")

    def __init__(
        self,
        client: tanjun.Client,
        interaction: hikari.CommandInteraction,
        register_task: collections.Callable[[asyncio.Task[typing.Any]], None],
        *,
        default_to_ephemeral: bool = False,
        future: typing.Optional[asyncio.Future[_ResponseTypeT]] = None,
        on_not_found: typing.Optional[collections.Callable[[tanjun.MenuContext], collections.Awaitable[None]]] = None,
    ) -> None:
        """Initialise a menu command context.

        Parameters
        ----------
        client
            The Tanjun client this context is bound to.
        interaction
            The command interaction this context is for.
        register_task
            Callback used to register long-running tasks spawned by this context.
        future
            A future used to set the initial response if this is being called
            through the REST webhook flow.
        default_to_ephemeral
            Whether to default to ephemeral responses.
        on_not_found
            Callback used to indicate no matching command was found.
        """
        super().__init__(client, interaction, register_task, default_to_ephemeral=default_to_ephemeral, future=future)
        self._command: typing.Optional[tanjun.MenuCommand[typing.Any, typing.Any]] = None
        self._marked_not_found = False
        self._on_not_found = on_not_found
        self._set_type_special_case(tanjun.MenuContext, self)._set_type_special_case(MenuContext, self)

    @property
    def command(self) -> typing.Optional[tanjun.MenuCommand[typing.Any, typing.Any]]:
        # <<inherited docstring from tanjun.abc.MenuContext>>.
        return self._command

    @property
    def target_id(self) -> hikari.Snowflake:
        # <<inherited docstring from tanjun.abc.MenuContext>>.
        assert self._interaction.resolved
        mapping = self._interaction.resolved.users or self._interaction.resolved.messages

        if not mapping:
            raise RuntimeError("Unknown menu type")

        return next(iter(mapping.keys()))

    @property
    def target(self) -> typing.Union[hikari.InteractionMember, hikari.User, hikari.Message]:
        # <<inherited docstring from tanjun.abc.MenuContext>>.
        assert self._interaction.resolved
        mapping = (
            self._interaction.resolved.members
            or self._interaction.resolved.users
            or self._interaction.resolved.messages
        )

        if not mapping:
            raise RuntimeError("Unknown menu type")

        return next(iter(mapping.values()))

    @property
    def triggering_name(self) -> str:
        # <<inherited docstring from tanjun.abc.Context>>.
        return self._interaction.command_name

    @property
    def type(self) -> typing.Literal[hikari.CommandType.USER, hikari.CommandType.MESSAGE]:
        # <<inherited docstring from tanjun.abc.MenuContext>>.
        command_type = hikari.CommandType(self._interaction.command_type)
        assert command_type in _VALID_TYPES
        return command_type

    async def mark_not_found(self) -> None:
        # <<inherited docstring from tanjun.abc.AppCommandContext>>.
        # TODO: assert not finalised?
        if self._on_not_found and not self._marked_not_found:
            self._marked_not_found = True
            await self._on_not_found(self)

    def set_command(self, command: typing.Optional[tanjun.MenuCommand[typing.Any, typing.Any]], /) -> Self:
        # <<inherited docstring from tanjun.abc.MenuContext>>.
        if command:
            self._set_type_special_case(tanjun.MenuCommand, command)

        elif self._command:
            self._remove_type_special_case(tanjun.MenuContext)

        self._command = command
        return self

    @typing.overload
    def resolve_to_member(self) -> hikari.InteractionMember:
        ...

    @typing.overload
    def resolve_to_member(self, *, default: _T) -> typing.Union[hikari.InteractionMember, _T]:
        ...

    def resolve_to_member(
        self, *, default: typing.Union[_T, _internal.Default] = _internal.DEFAULT
    ) -> typing.Union[hikari.InteractionMember, _T]:
        # <<inherited docstring from tanjun.abc.MenuContext>>.
        assert self._interaction.resolved
        if self._interaction.resolved.members:
            return next(iter(self._interaction.resolved.members.values()))

        if self._interaction.resolved.users:
            if default is not _internal.DEFAULT:
                return default

            raise LookupError("User isn't in the current guild")

        raise TypeError("Cannot resolve message menu context to a user")

    def resolve_to_message(self) -> hikari.Message:
        # <<inherited docstring from tanjun.abc.MenuContext>>.
        assert self._interaction.resolved
        if self._interaction.resolved.messages:
            return next(iter(self._interaction.resolved.messages.values()))

        raise TypeError("Cannot resolve user menu context to a message")

    def resolve_to_user(self) -> typing.Union[hikari.User, hikari.Member]:
        # <<inherited docstring from tanjun.abc.MenuContext>>.
        assert self._interaction.resolved
        return self.resolve_to_member(default=None) or next(iter(self._interaction.resolved.users.values()))



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportPrivateUsage=none
# This leads to too many false-positives around mocks.

import inspect
import typing
from unittest import mock

import hikari
import pytest

import tanjun
from tanjun import _internal


@pytest.mark.asyncio()
async def test_gather_checks_handles_no_checks():
    mock_ctx = mock.AsyncMock()
    assert await _internal.gather_checks(mock_ctx, ()) is True

    mock_ctx.call_with_async_di.assert_not_called()


@pytest.mark.asyncio()
async def test_gather_checks_handles_failed_check():
    mock_ctx = mock.Mock()
    mock_ctx.call_with_async_di = mock.AsyncMock(side_effect=[True, False, True])
    check_1 = mock.Mock()
    check_2 = mock.Mock()
    check_3 = mock.Mock()

    assert await _internal.gather_checks(mock_ctx, (check_1, check_2, check_3)) is False

    mock_ctx.call_with_async_di.assert_has_awaits(
        [mock.call(check_1, mock_ctx), mock.call(check_2, mock_ctx), mock.call(check_3, mock_ctx)]
    )


@pytest.mark.asyncio()
async def test_gather_checks_handles_check_failed_by_raise():
    mock_ctx = mock.Mock()
    mock_ctx.call_with_async_di = mock.AsyncMock(side_effect=[True, tanjun.FailedCheck, True])
    check_1 = mock.Mock()
    check_2 = mock.Mock()
    check_3 = mock.Mock()

    assert await _internal.gather_checks(mock_ctx, (check_1, check_2, check_3)) is False

    mock_ctx.call_with_async_di.assert_has_awaits(
        [mock.call(check_1, mock_ctx), mock.call(check_2, mock_ctx), mock.call(check_3, mock_ctx)]
    )


@pytest.mark.asyncio()
async def test_gather_checks():
    mock_ctx = mock.Mock()
    mock_ctx.call_with_async_di = mock.AsyncMock(side_effect=[True, True, True])
    check_1 = mock.Mock()
    check_2 = mock.Mock()
    check_3 = mock.Mock()

    assert await _internal.gather_checks(mock_ctx, (check_1, check_2, check_3)) is True

    mock_ctx.call_with_async_di.assert_has_awaits(
        [mock.call(check_1, mock_ctx), mock.call(check_2, mock_ctx), mock.call(check_3, mock_ctx)]
    )


class TestCastedView:
    def test___getitem___for_non_existant_entry(self):
        mock_cast = mock.Mock()
        view = _internal.CastedView[typing.Any, typing.Any]({}, mock_cast)

        with pytest.raises(KeyError):
            view["foo"]

        mock_cast.assert_not_called()

    def test___getitem___for_buffered_entry(self):
        mock_cast = mock.Mock()
        mock_value = mock.MagicMock()
        view = _internal.CastedView[typing.Any, str]({"a": "b"}, mock_cast)
        view._buffer["a"] = mock_value

        result = view["a"]

        assert result is mock_value
        mock_cast.assert_not_called()

    def test___getitem___for_not_buffered_entry(self):
        mock_cast = mock.Mock()
        mock_value = mock.MagicMock()
        view = _internal.CastedView[typing.Any, typing.Any]({"a": mock_value}, mock_cast)

        result = view["a"]

        assert result is mock_cast.return_value
        assert view._buffer["a"] is mock_cast.return_value
        mock_cast.assert_called_once_with(mock_value)

    def test___iter__(self):
        mock_iter = mock.Mock(return_value=iter((1, 2, 3)))
        mock_dict = mock.Mock(__iter__=mock_iter)
        view = _internal.CastedView[int, int](mock_dict, mock.Mock())

        assert iter(view) is mock_iter.return_value

    def test___len___(self):
        mock_dict = mock.Mock(__len__=mock.Mock(return_value=43123))
        view = _internal.CastedView[int, int](mock_dict, mock.Mock())

        assert len(view) == 43123


def test_ensure_parse_channel_types_has_every_channel_class():
    for _, attribute in inspect.getmembers(hikari):
        if isinstance(attribute, type) and issubclass(attribute, hikari.PartialChannel):
            result = _internal.parse_channel_types(attribute)

            assert result


def test_ensure_repr_channel_has_every_real_channel_type():
    for channel_type in hikari.ChannelType:
        result = _internal.repr_channel(channel_type)

        assert isinstance(result, str)
        assert result
        assert result != "Unknown"


def test_ensure_repr_channel_defaults_to_unknown():
    assert _internal.repr_channel(hikari.ChannelType(-1)) == "Unknown"



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Menu command implementations."""
from __future__ import annotations

__all__: list[str] = ["MenuCommand", "as_message_menu", "as_user_menu"]

import typing

import hikari

from .. import _internal
from .. import abc as tanjun
from .. import components
from .. import errors
from .. import hooks as hooks_
from .._internal import localisation
from . import base

if typing.TYPE_CHECKING:
    from collections import abc as collections

    from typing_extensions import Self

    _AnyCallbackSigT = typing.TypeVar(
        "_AnyCallbackSigT", bound=collections.Callable[..., collections.Coroutine[typing.Any, typing.Any, None]]
    )
    _MessageCallbackSigT = typing.TypeVar("_MessageCallbackSigT", bound=tanjun.MenuCallbackSig[hikari.Message])
    _UserCallbackSigT = typing.TypeVar("_UserCallbackSigT", bound=tanjun.MenuCallbackSig[hikari.InteractionMember])

    _AnyCommandT = typing.Union[
        tanjun.MenuCommand[_AnyCallbackSigT, typing.Any],
        tanjun.MessageCommand[_AnyCallbackSigT],
        tanjun.SlashCommand[_AnyCallbackSigT],
    ]
    _CallbackishT = typing.Union[_AnyCallbackSigT, _AnyCommandT[_AnyCallbackSigT]]

_AnyMenuCallbackSigT = typing.TypeVar("_AnyMenuCallbackSigT", bound=tanjun.MenuCallbackSig[typing.Any])
_MenuTypeT = typing.TypeVar(
    "_MenuTypeT", typing.Literal[hikari.CommandType.USER], typing.Literal[hikari.CommandType.MESSAGE]
)
_EMPTY_HOOKS: typing.Final[hooks_.Hooks[typing.Any]] = hooks_.Hooks()


# While these overloads may seem redundant/unnecessary, MyPy cannot understand
# this when expressed through `callback: _CallbackIshT[_MessageCallbackSigT]`.
class _AsMsgResultProto(typing.Protocol):
    @typing.overload
    def __call__(
        self, _: _MessageCallbackSigT, /
    ) -> MenuCommand[_MessageCallbackSigT, typing.Literal[hikari.CommandType.MESSAGE]]:
        ...

    @typing.overload
    def __call__(
        self, _: _AnyCommandT[_MessageCallbackSigT], /
    ) -> MenuCommand[_MessageCallbackSigT, typing.Literal[hikari.CommandType.MESSAGE]]:
        ...


def as_message_menu(
    name: typing.Union[str, collections.Mapping[str, str]],
    /,
    *,
    always_defer: bool = False,
    default_member_permissions: typing.Union[hikari.Permissions, int, None] = None,
    default_to_ephemeral: typing.Optional[bool] = None,
    dm_enabled: typing.Optional[bool] = None,
    is_global: bool = True,
) -> _AsMsgResultProto:
    """Build a message [tanjun.MenuCommand][] by decorating a function.

    !!! note
        Under the standard implementation, `is_global` is used to determine whether
        the command should be bulk set by [tanjun.Client.declare_global_commands][]
        or when `declare_global_commands` is True

    !!! note
        If you want your first response to be ephemeral while using
        `always_defer`, you must set `default_to_ephemeral` to `True`.

    Examples
    --------
    ```py
    @as_message_menu("message")
    async def message_command(self, ctx: tanjun.abc.MenuContext, message: hikari.Message) -> None:
        await ctx.respond(
            embed=hikari.Embed(title="Message content", description=message.content or "N/A")
        )
    ```

    Parameters
    ----------
    name
        The command's name (supports [localisation][]).

        This must be between 1 and 32 characters in length.
    always_defer
        Whether the contexts this command is executed with should always be deferred
        before being passed to the command's callback.
    default_member_permissions
        Member permissions necessary to utilize this command by default.

        If this is [None][] then the configuration for the parent component or client
        will be used.
    default_to_ephemeral
        Whether this command's responses should default to ephemeral unless flags
        are set to override this.

        If this is left as [None][] then the default set on the parent command(s),
        component or client will be in effect.
    dm_enabled
        Whether this command is enabled in DMs with the bot.

        If this is [None][] then the configuration for the parent component or client
        will be used.
    is_global
        Whether this command is a global command.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.MenuCallbackSig], MenuCommand]
        The decorator callback used to make a [tanjun.MenuCommand][].

        This can either wrap a raw command callback or another callable command instance
        (e.g. [tanjun.MenuCommand][], [tanjun.MessageCommand][], [tanjun.SlashCommand][])
        and will manage loading the other command into a component when using
        [tanjun.Component.load_from_scope][].

    Raises
    ------
    ValueError
        Raises a value error for any of the following reasons:

        * If the command name isn't in the length range of 1 to 32.
        * If the command name has uppercase characters.
    """

    def decorator(
        callback: _CallbackishT[_MessageCallbackSigT], /
    ) -> MenuCommand[_MessageCallbackSigT, typing.Literal[hikari.CommandType.MESSAGE]]:
        if isinstance(callback, (tanjun.MenuCommand, tanjun.MessageCommand, tanjun.SlashCommand)):
            wrapped_command = callback
            callback = callback.callback

        else:
            wrapped_command = None

        return MenuCommand(
            callback,
            hikari.CommandType.MESSAGE,
            name,
            always_defer=always_defer,
            default_member_permissions=default_member_permissions,
            default_to_ephemeral=default_to_ephemeral,
            dm_enabled=dm_enabled,
            is_global=is_global,
            _wrapped_command=wrapped_command,
        )

    return decorator


# While these overloads may seem redundant/unnecessary, MyPy cannot understand
# this when expressed through `callback: _CallbackIshT[_MessageCallbackSigT]`.
class _AsUserResultProto(typing.Protocol):
    @typing.overload
    def __call__(
        self, _: _UserCallbackSigT, /
    ) -> MenuCommand[_UserCallbackSigT, typing.Literal[hikari.CommandType.USER]]:
        ...

    @typing.overload
    def __call__(
        self, _: _AnyCommandT[_UserCallbackSigT], /
    ) -> MenuCommand[_UserCallbackSigT, typing.Literal[hikari.CommandType.USER]]:
        ...


def as_user_menu(
    name: typing.Union[str, collections.Mapping[str, str]],
    /,
    *,
    always_defer: bool = False,
    default_member_permissions: typing.Union[hikari.Permissions, int, None] = None,
    default_to_ephemeral: typing.Optional[bool] = None,
    dm_enabled: typing.Optional[bool] = None,
    is_global: bool = True,
) -> _AsUserResultProto:
    """Build a user [tanjun.MenuCommand][] by decorating a function.

    !!! note
        Under the standard implementation, `is_global` is used to determine whether
        the command should be bulk set by [tanjun.Client.declare_global_commands][]
        or when `declare_global_commands` is True

    !!! note
        If you want your first response to be ephemeral while using
        `always_defer`, you must set `default_to_ephemeral` to `True`.

    Examples
    --------
    ```py
    @as_user_menu("user")
    async def user_command(
        self,
        ctx: tanjun.abc.MenuContext,
        user: hikari.User | hikari.InteractionMember,
    ) -> None:
        await ctx.respond(f"Hello {user}")
    ```

    Parameters
    ----------
    name
        The command's name (supports [localisation][]).

        This must be between 1 and 32 characters in length.
    always_defer
        Whether the contexts this command is executed with should always be deferred
        before being passed to the command's callback.
    default_member_permissions
        Member permissions necessary to utilize this command by default.

        If this is [None][] then the configuration for the parent component or client
        will be used.
    default_to_ephemeral
        Whether this command's responses should default to ephemeral unless flags
        are set to override this.

        If this is left as [None][] then the default set on the parent command(s),
        component or client will be in effect.
    dm_enabled
        Whether this command is enabled in DMs with the bot.

        If this is [None][] then the configuration for the parent component or client
        will be used.
    is_global
        Whether this command is a global command.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.MenuCallbackSig], MenuCommand]
        The decorator callback used to make a [tanjun.MenuCommand][].

        This can either wrap a raw command callback or another callable command instance
        (e.g. [tanjun.MenuCommand][], [tanjun.MessageCommand][], [tanjun.SlashCommand][])
        and will manage loading the other command into a component when using
        [tanjun.Component.load_from_scope][].

    Raises
    ------
    ValueError
        Raises a value error for any of the following reasons:

        * If the command name isn't in the length range of 1 to 32.
        * If the command name has uppercase characters.
    """

    def decorator(
        callback: _CallbackishT[_UserCallbackSigT], /
    ) -> MenuCommand[_UserCallbackSigT, typing.Literal[hikari.CommandType.USER]]:
        if isinstance(callback, (tanjun.MenuCommand, tanjun.MessageCommand, tanjun.SlashCommand)):
            wrapped_command = callback
            callback = callback.callback

        else:
            wrapped_command = None

        return MenuCommand(
            callback,
            hikari.CommandType.USER,
            name,
            always_defer=always_defer,
            default_member_permissions=default_member_permissions,
            default_to_ephemeral=default_to_ephemeral,
            dm_enabled=dm_enabled,
            is_global=is_global,
            _wrapped_command=wrapped_command,
        )

    return decorator


_VALID_TYPES = frozenset((hikari.CommandType.MESSAGE, hikari.CommandType.USER))


class MenuCommand(base.PartialCommand[tanjun.MenuContext], tanjun.MenuCommand[_AnyMenuCallbackSigT, _MenuTypeT]):
    """Base class used for the standard menu command implementations."""

    __slots__ = (
        "_always_defer",
        "_callback",
        "_default_member_permissions",
        "_defaults_to_ephemeral",
        "_description",
        "_is_dm_enabled",
        "_is_global",
        "_names",
        "_parent",
        "_tracked_command",
        "_type",
        "_wrapped_command",
    )

    @typing.overload
    def __init__(
        self: MenuCommand[_UserCallbackSigT, typing.Literal[hikari.CommandType.USER]],
        callback: _UserCallbackSigT,
        type_: typing.Literal[hikari.CommandType.USER],
        name: typing.Union[str, collections.Mapping[str, str]],
        /,
        *,
        always_defer: bool = False,
        default_member_permissions: typing.Union[hikari.Permissions, int, None] = None,
        default_to_ephemeral: typing.Optional[bool] = None,
        dm_enabled: typing.Optional[bool] = None,
        is_global: bool = True,
        _wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]] = None,
    ) -> None:
        ...

    # While this extra overload may seem redundant/unnecessary, MyPy cannot understand
    # this when expressed through `callback: _CallbackIshT[_MessageCallbackSigT]`.
    @typing.overload
    def __init__(
        self: MenuCommand[_UserCallbackSigT, typing.Literal[hikari.CommandType.USER]],
        callback: _AnyCommandT[_UserCallbackSigT],
        type_: typing.Literal[hikari.CommandType.USER],
        name: typing.Union[str, collections.Mapping[str, str]],
        /,
        *,
        always_defer: bool = False,
        default_member_permissions: typing.Union[hikari.Permissions, int, None] = None,
        default_to_ephemeral: typing.Optional[bool] = None,
        dm_enabled: typing.Optional[bool] = None,
        is_global: bool = True,
        _wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]] = None,
    ) -> None:
        ...

    @typing.overload
    def __init__(
        self: MenuCommand[_MessageCallbackSigT, typing.Literal[hikari.CommandType.MESSAGE]],
        callback: _MessageCallbackSigT,
        type_: typing.Literal[hikari.CommandType.MESSAGE],
        name: typing.Union[str, collections.Mapping[str, str]],
        /,
        *,
        always_defer: bool = False,
        default_member_permissions: typing.Union[hikari.Permissions, int, None] = None,
        default_to_ephemeral: typing.Optional[bool] = None,
        dm_enabled: typing.Optional[bool] = None,
        is_global: bool = True,
        _wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]] = None,
    ) -> None:
        ...

    # While this extra overload may seem redundant/unnecessary, MyPy cannot understand
    # this when expressed through `callback: _CallbackIshT[_MessageCallbackSigT]`.
    @typing.overload
    def __init__(
        self: MenuCommand[_MessageCallbackSigT, typing.Literal[hikari.CommandType.MESSAGE]],
        callback: _AnyCommandT[_MessageCallbackSigT],
        type_: typing.Literal[hikari.CommandType.MESSAGE],
        name: typing.Union[str, collections.Mapping[str, str]],
        /,
        *,
        always_defer: bool = False,
        default_member_permissions: typing.Union[hikari.Permissions, int, None] = None,
        default_to_ephemeral: typing.Optional[bool] = None,
        dm_enabled: typing.Optional[bool] = None,
        is_global: bool = True,
        _wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]] = None,
    ) -> None:
        ...

    def __init__(
        self,
        callback: _CallbackishT[_AnyMenuCallbackSigT],
        type_: _MenuTypeT,
        name: typing.Union[str, collections.Mapping[str, str]],
        /,
        *,
        always_defer: bool = False,
        default_member_permissions: typing.Union[hikari.Permissions, int, None] = None,
        default_to_ephemeral: typing.Optional[bool] = None,
        dm_enabled: typing.Optional[bool] = None,
        is_global: bool = True,
        _wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]] = None,
    ) -> None:
        """Initialise a user or message menu command.

        !!! note
            Under the standard implementation, `is_global` is used to determine whether
            the command should be bulk set by [tanjun.Client.declare_global_commands][]
            or when `declare_global_commands` is True

        !!! note
            If you want your first response to be ephemeral while using
            `always_defer`, you must set `default_to_ephemeral` to `True`.

        Parameters
        ----------
        callback : tanjun.abc.MenuCallbackSig
            Callback to execute when the command is invoked.

            This should be an asynchronous callback which takes one positional
            argument of type [tanjun.abc.MenuContext][], returns [None][]
            and may use dependency injection to access other services.
        type_ : hikari.commands.CommandType
            The type of menu command this is.

            Only [hikari.commands.CommandType.USER][] and [hikari.commands.CommandType.MESSAGE][]
            are valid here.
        name
            The command's name (supports [localisation][]).

            This must be between 1 and 32 characters in length.
        always_defer
            Whether the contexts this command is executed with should always be deferred
            before being passed to the command's callback.
        default_member_permissions
            Member permissions necessary to utilize this command by default.

            If this is [None][] then the configuration for the parent component or client
            will be used.
        default_to_ephemeral
            Whether this command's responses should default to ephemeral unless flags
            are set to override this.

            If this is left as [None][] then the default set on the parent command(s),
            component or client will be in effect.
        dm_enabled
            Whether this command is enabled in DMs with the bot.

            If this is [None][] then the configuration for the parent component or client
            will be used.
        is_global
            Whether this command is a global command.

        Returns
        -------
        collections.abc.Callable[[tanjun.abc.MenuCallbackSig], MenuCommand]
            The decorator callback used to make a [tanjun.MenuCommand][].

            This can either wrap a raw command callback or another callable command instance
            (e.g. [tanjun.MenuCommand][], [tanjun.MessageCommand][], [tanjun.SlashCommand][])
            and will manage loading the other command into a component when using
            [tanjun.Component.load_from_scope][].

        Raises
        ------
        ValueError
            Raises a value error for any of the following reasons:

            * If the command name isn't in the length range of 1 to 32.
            * If the command name has uppercase characters.
        """
        super().__init__()
        names = localisation.MaybeLocalised("name", name).assert_length(1, 32)

        if type_ not in _VALID_TYPES:
            raise ValueError("Command type must be message or user")

        if isinstance(callback, (tanjun.MenuCommand, tanjun.MessageCommand, tanjun.SlashCommand)):
            callback = callback.callback

        if default_member_permissions is not None:
            default_member_permissions = hikari.Permissions(default_member_permissions)

        self._always_defer = always_defer
        self._callback: _AnyMenuCallbackSigT = callback
        self._default_member_permissions = default_member_permissions
        self._defaults_to_ephemeral = default_to_ephemeral
        self._is_dm_enabled = dm_enabled
        self._is_global = is_global
        self._names = names
        self._parent: typing.Optional[tanjun.SlashCommandGroup] = None
        self._tracked_command: typing.Optional[hikari.ContextMenuCommand] = None
        self._type: _MenuTypeT = type_  # MyPy bug causes this to need an explicit annotation.
        self._wrapped_command = _wrapped_command

    if typing.TYPE_CHECKING:
        __call__: _AnyMenuCallbackSigT  # pyright: ignore [ reportUninitializedInstanceVariable ]

    else:

        async def __call__(self, *args, **kwargs) -> None:
            await self._callback(*args, **kwargs)

    @property
    def callback(self) -> _AnyMenuCallbackSigT:
        # <<inherited docstring from tanjun.abc.MenuCommand>>.
        return self._callback

    @property
    def default_member_permissions(self) -> typing.Optional[hikari.Permissions]:
        # <<inherited docstring from tanjun.abc.AppCommand>>.
        return self._default_member_permissions

    @property
    def defaults_to_ephemeral(self) -> typing.Optional[bool]:
        # <<inherited docstring from tanjun.abc.MenuCommand>>.
        return self._defaults_to_ephemeral

    @property
    def is_dm_enabled(self) -> typing.Optional[bool]:
        return self._is_dm_enabled

    @property
    def is_global(self) -> bool:
        # <<inherited docstring from tanjun.abc.AppCommand>>.
        return self._is_global

    @property
    def name(self) -> str:
        # <<inherited docstring from tanjun.abc.AppCommand>>.
        return self._names.default_value

    @property
    def name_localisations(self) -> collections.Mapping[str, str]:
        return self._names.localised_values.copy()

    @property
    def tracked_command(self) -> typing.Optional[hikari.ContextMenuCommand]:
        # <<inherited docstring from tanjun.abc.MenuCommand>>.
        return self._tracked_command

    @property
    def tracked_command_id(self) -> typing.Optional[hikari.Snowflake]:
        # <<inherited docstring from tanjun.abc.AppCommand>>.
        return self._tracked_command.id if self._tracked_command else None

    @property
    def type(self) -> _MenuTypeT:
        # <<inherited docstring from tanjun.abc.AppCommand>>.
        return self._type

    @property
    def wrapped_command(self) -> typing.Optional[tanjun.ExecutableCommand[typing.Any]]:
        """The command object this wraps, if any."""
        return self._wrapped_command

    def build(self, *, component: typing.Optional[tanjun.Component] = None) -> hikari.api.ContextMenuCommandBuilder:
        # <<inherited docstring from tanjun.abc.MenuCommand>>.
        builder = hikari.impl.ContextMenuCommandBuilder(
            type=self._type, name=self._names.default_value, name_localizations=self._names.localised_values
        )

        component = component or self._component
        if self._default_member_permissions is not None:
            builder.set_default_member_permissions(self._default_member_permissions)
        elif component and component.default_app_cmd_permissions is not None:
            builder.set_default_member_permissions(component.default_app_cmd_permissions)

        if self._is_dm_enabled is not None:
            builder.set_is_dm_enabled(self._is_dm_enabled)
        elif component and component.dms_enabled_for_app_cmds is not None:
            builder.set_is_dm_enabled(component.dms_enabled_for_app_cmds)

        return builder

    def set_tracked_command(self, command: hikari.PartialCommand, /) -> Self:
        # <<inherited docstring from tanjun.abc.MenuCommand>>.
        if not isinstance(command, hikari.ContextMenuCommand):
            raise TypeError("Command must be a ContextMenuCommand")

        self._tracked_command = command
        return self

    def set_ephemeral_default(self, state: typing.Optional[bool], /) -> Self:
        """Set whether this command's responses should default to ephemeral.

        Parameters
        ----------
        state
            Whether this command's responses should default to ephemeral.
            This will be overridden by any response calls which specify flags.

            Setting this to [None][] will let the default set on the parent
            command(s), component or client propagate and decide the ephemeral
            default for contexts used by this command.

        Returns
        -------
        Self
            This command to allow for chaining.
        """
        self._defaults_to_ephemeral = state
        return self

    async def check_context(self, ctx: tanjun.MenuContext, /) -> bool:
        # <<inherited docstring from tanjun.abc.MenuCommand>>.
        ctx.set_command(self)
        result = await _internal.gather_checks(ctx, self._checks)
        ctx.set_command(None)
        return result

    def copy(self, *, parent: typing.Optional[tanjun.SlashCommandGroup] = None) -> Self:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        inst = super().copy()
        inst._parent = parent
        return inst

    async def execute(
        self, ctx: tanjun.MenuContext, /, *, hooks: typing.Optional[collections.MutableSet[tanjun.MenuHooks]] = None
    ) -> None:
        # <<inherited docstring from tanjun.abc.MenuCommand>>.
        if self._always_defer and not ctx.has_been_deferred and not ctx.has_responded:
            await ctx.defer()

        ctx = ctx.set_command(self)
        own_hooks = self._hooks or _EMPTY_HOOKS
        try:
            await own_hooks.trigger_pre_execution(ctx, hooks=hooks)

            if self._type is hikari.CommandType.USER:
                value: typing.Union[hikari.Message, hikari.User] = ctx.resolve_to_user()

            else:
                value = ctx.resolve_to_message()

            await ctx.call_with_async_di(self._callback, ctx, value)

        except errors.CommandError as exc:
            await exc.send(ctx)

        except errors.HaltExecution:
            # Unlike a message command, this won't necessarily reach the client level try except
            # block so we have to handle this here.
            await ctx.mark_not_found()

        except Exception as exc:
            if await own_hooks.trigger_error(ctx, exc, hooks=hooks) <= 0:
                raise

        else:
            await own_hooks.trigger_success(ctx, hooks=hooks)

        finally:
            await own_hooks.trigger_post_execution(ctx, hooks=hooks)

    def load_into_component(self, component: tanjun.Component, /) -> None:
        # <<inherited docstring from tanjun.components.load_into_component>>.
        component.add_menu_command(self)
        if self._wrapped_command and isinstance(self._wrapped_command, components.AbstractComponentLoader):
            self._wrapped_command.load_into_component(component)



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Functions and classes used to enable more Discord oriented argument converters."""
from __future__ import annotations

__all__: list[str] = [
    "ToChannel",
    "ToEmoji",
    "ToGuild",
    "ToInvite",
    "ToInviteWithMetadata",
    "ToMember",
    "ToMessage",
    "ToPresence",
    "ToRole",
    "ToUser",
    "ToVoiceState",
    "from_datetime",
    "parse_channel_id",
    "parse_emoji_id",
    "parse_message_id",
    "parse_role_id",
    "parse_snowflake",
    "parse_user_id",
    "search_channel_ids",
    "search_emoji_ids",
    "search_role_ids",
    "search_snowflakes",
    "search_user_ids",
    "to_bool",
    "to_channel",
    "to_color",
    "to_colour",
    "to_datetime",
    "to_emoji",
    "to_guild",
    "to_invite",
    "to_invite_with_metadata",
    "to_member",
    "to_message",
    "to_presence",
    "to_role",
    "to_snowflake",
    "to_user",
    "to_voice_state",
]

import datetime
import functools
import logging
import operator
import re
import typing
import urllib.parse as urlparse
from collections import abc as collections

import alluka
import hikari
import typing_extensions

from . import _internal
from . import abc as tanjun
from .dependencies import async_cache

if typing.TYPE_CHECKING:
    from . import parsing

    _PartialChannelT = typing.TypeVar("_PartialChannelT", bound=hikari.PartialChannel)


_SnowflakeIsh = typing.Union[str, int]
_ValueT = typing.TypeVar("_ValueT")
_LOGGER = logging.getLogger("hikari.tanjun.conversion")


class BaseConverter:
    """Base class for the standard converters.

    !!! warning
        Inheriting from this is completely unnecessary and should be avoided
        for people using the library unless they know what they're doing.

    This is detail of the standard implementation and isn't guaranteed to work
    between implementations but will work for implementations which provide
    the standard dependency injection or special cased support for these.

    While it isn't necessary to subclass this to implement your own converters
    since dependency injection can be used to access fields like the current Context,
    this class introduces some niceties around stuff like state warnings.
    """

    __slots__ = ("__weakref__",)

    @property
    @typing_extensions.deprecated("Use .caches instead")
    def async_caches(self) -> collections.Sequence[typing.Any]:
        """Deprecated property."""
        return list({v[0] for v in self.caches})

    @property
    @typing_extensions.deprecated("Use .caches instead")
    def cache_components(self) -> hikari.api.CacheComponents:
        """Deprecated property."""
        if self.caches:
            return functools.reduce(operator.ior, (v[1] for v in self.caches))

        return hikari.api.CacheComponents.NONE

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        """Caches the converter takes advantage of.

        This returns a tuple of async cache types and the relevant cache components
        which will be needed if said async cache isn't implemented.

        !!! note
            Unless [tanjun.conversion.BaseConverter.requires_cache][] is [True][],
            these cache components aren't necessary but simply avoid the converter
            from falling back to REST requests.
        """
        return []

    @property
    @typing_extensions.deprecated("Use .caches instead")
    def intents(self) -> hikari.Intents:
        """Deprecated property."""
        if self.caches:
            return functools.reduce(operator.ior, (v[2] for v in self.caches))

        return hikari.Intents.NONE

    @property
    def requires_cache(self) -> bool:
        """Whether this converter relies on the relevant cache stores to work.

        If this is [True][] then this converter will not function properly
        in an environment [tanjun.conversion.BaseConverter.intents][] or
        [tanjun.conversion.BaseConverter.cache_components][] isn't satisfied
        and will never fallback to REST requests.
        """
        return False

    def check_client(self, client: tanjun.Client, parent_name: str, /) -> None:
        """Check that this converter will work with the given client.

        This never raises any errors but simply warns the user if the converter
        is not compatible with the given client.

        Parameters
        ----------
        client
            The client to check against.
        parent_name
            The name of the converter's parent, used for warning messages.
        """
        enabled_components = client.cache and client.cache.settings.components or hikari.api.CacheComponents.NONE
        enabled_intents = client.shards.intents if client.shards else hikari.Intents.NONE
        missing_components = hikari.api.CacheComponents.NONE
        needed_intents = hikari.Intents.NONE
        for cache_dep_type, component, intents in self.caches:
            if client.injector.get_type_dependency(cache_dep_type, default=None) is not None:
                continue

            needed_intents |= intents
            if current_missing := component & ~enabled_components:
                missing_components |= current_missing

        if missing_components:
            if self.requires_cache:
                _LOGGER.warning(
                    "Converter %r (registered with %s) will fail without the following cache components: %s",
                    self,
                    parent_name,
                    missing_components,
                )

            else:
                _LOGGER.info(
                    "Converter %r (registered with %s) may not perform "
                    "optimally without the following cache components: %s",
                    self,
                    parent_name,
                    missing_components,
                )

        if missing_intents := needed_intents & ~enabled_intents:
            if self.requires_cache:
                _LOGGER.warning(
                    "Converter %r (registered with %s) will fail without the following intents: %s",
                    self,
                    parent_name,
                    missing_intents,
                )

            else:
                _LOGGER.info(
                    "Converter %r (registered with %s) may not perform as expected without the following intents: %s",
                    self,
                    parent_name,
                    missing_intents,
                )


_DmCacheT = async_cache.SfCache[hikari.DMChannel]
_GuildChannelCacheT = async_cache.SfCache[hikari.PermissibleGuildChannel]
_ThreadCacheT = async_cache.SfCache[hikari.GuildThreadChannel]


class ToChannel(BaseConverter):
    """Standard converter for channels mentions/IDs.

    For a standard instance of this see [tanjun.conversion.to_channel][].
    """

    __slots__ = ("_allowed_types", "_allowed_types_repr", "_dms_enabled", "_guilds_enabled", "_threads_enabled")

    def __init__(
        self,
        *,
        allowed_types: typing.Optional[collections.Collection[typing.Union[type[hikari.PartialChannel], int]]] = None,
        include_dms: bool = True,
    ) -> None:
        """Initialise a to channel converter.

        Parameters
        ----------
        allowed_types
            Collection of channel types and classes to allow.

            If this is [None][] then all channel types will be allowed.
        include_dms
            Whether to include DM channels in the results.

            May lead to a lot of extra fallbacks to REST requests if
            the client doesn't have a registered async cache for DMs.
        """
        allowed_types_ = _internal.parse_channel_types(*allowed_types or ())
        dm_types = _internal.CHANNEL_TYPES[hikari.PrivateChannel]

        if not include_dms:
            if allowed_types is None:
                allowed_types = _internal.CHANNEL_TYPES[hikari.PartialChannel].difference(dm_types)
                allowed_types_ = list(allowed_types)

            else:
                for channel_type in dm_types:
                    try:
                        allowed_types_.remove(channel_type)
                    except ValueError:
                        pass

        self._allowed_types = set(allowed_types_) if allowed_types is not None else None
        self._dms_enabled = include_dms and (
            self._allowed_types is None or any(map(self._allowed_types.__contains__, dm_types))
        )
        self._guilds_enabled = self._allowed_types is None or any(
            map(self._allowed_types.__contains__, _internal.CHANNEL_TYPES[hikari.PermissibleGuildChannel])
        )
        self._threads_enabled = self._allowed_types is None or any(
            map(self._allowed_types.__contains__, _internal.CHANNEL_TYPES[hikari.GuildThreadChannel])
        )

        if not allowed_types_:
            self._allowed_types_repr = ""

        elif len(allowed_types_) == 1:
            self._allowed_types_repr = _internal.repr_channel(allowed_types_[0])

        else:
            self._allowed_types_repr = ", ".join(map(_internal.repr_channel, allowed_types_[:-1]))
            self._allowed_types_repr += f" and {_internal.repr_channel(allowed_types_[-1])}"

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        # <<inherited docstring from BaseConverter>>.
        results: list[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]] = []

        if self._guilds_enabled:
            results.append((_GuildChannelCacheT, hikari.api.CacheComponents.GUILD_CHANNELS, hikari.Intents.GUILDS))

        if self._dms_enabled:
            results.append((_DmCacheT, hikari.api.CacheComponents.NONE, hikari.Intents.NONE))

        if self._threads_enabled:
            results.append((_ThreadCacheT, hikari.api.CacheComponents.NONE, hikari.Intents.GUILDS))

        return results

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return False

    def _assert_type(self, channel: _PartialChannelT, /) -> _PartialChannelT:
        if self._allowed_types is not None and channel.type not in self._allowed_types:
            raise ValueError(
                f"Only the following channel types are allowed for this argument: {self._allowed_types_repr}"
            )

        return channel

    async def __call__(
        self,
        argument: _SnowflakeIsh,
        /,
        ctx: alluka.Injected[tanjun.Context],
        *,
        cache: alluka.Injected[typing.Optional[_GuildChannelCacheT]] = None,
        dm_cache: alluka.Injected[typing.Optional[_DmCacheT]] = None,
        thread_cache: alluka.Injected[typing.Optional[_ThreadCacheT]] = None,
    ) -> hikari.PartialChannel:
        channel_id = parse_channel_id(argument, message="No valid channel mention or ID found")
        if ctx.cache and (channel_ := ctx.cache.get_guild_channel(channel_id)):
            return self._assert_type(channel_)

        # Ensure bool for MyPy compat
        no_guild_channel = bool(
            (cache or not self._guilds_enabled)
            and (thread_cache or not self._threads_enabled)
            and (dm_cache or not self._dms_enabled)
        )
        if cache:
            try:
                return self._assert_type(await cache.get(channel_id))

            except async_cache.EntryNotFound:
                pass

            except async_cache.CacheMissError:
                no_guild_channel = False

        if thread_cache and self._threads_enabled:
            try:
                return self._assert_type(await thread_cache.get(channel_id))

            except async_cache.EntryNotFound:
                pass

            except async_cache.CacheMissError:
                no_guild_channel = False

        if dm_cache and self._dms_enabled:
            try:
                return self._assert_type(await dm_cache.get(channel_id))

            except async_cache.EntryNotFound:
                pass

            except async_cache.CacheMissError:
                no_guild_channel = False

        if not no_guild_channel:
            try:
                return self._assert_type(await ctx.rest.fetch_channel(channel_id))

            except hikari.NotFoundError:
                pass

        raise ValueError("Couldn't find channel")


@typing_extensions.deprecated("Use ToChannel instead")
class ChannelConverter(ToChannel):
    """Deprecated alias of [tanjun.conversion.ToChannel][]."""

    __slots__ = ()


_EmojiCacheT = async_cache.SfCache[hikari.KnownCustomEmoji]


class ToEmoji(BaseConverter):
    """Standard converter for custom emojis.

    For a standard instance of this see [tanjun.conversion.to_emoji][].

    !!! note
        If you just want to convert inpute to a [hikari.emojis.Emoji][],
        [hikari.emojis.CustomEmoji][] or [hikari.emojis.UnicodeEmoji][] without
        making any cache or REST calls then you can just use the relevant
        [Emoji.parse][hikari.emojis.Emoji.parse],
        [CustomEmoji.parse][hikari.emojis.CustomEmoji.parse] or
        [UnicodeEmoji.parse][hikari.emojis.UnicodeEmoji.parse] methods.
    """

    __slots__ = ()

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        # <<inherited docstring from BaseConverter>>.
        return [(_EmojiCacheT, hikari.api.CacheComponents.EMOJIS, hikari.Intents.GUILD_EMOJIS | hikari.Intents.GUILDS)]

    async def __call__(
        self,
        argument: _SnowflakeIsh,
        /,
        ctx: alluka.Injected[tanjun.Context],
        *,
        cache: alluka.Injected[typing.Optional[_EmojiCacheT]] = None,
    ) -> hikari.KnownCustomEmoji:
        emoji_id = parse_emoji_id(argument, message="No valid emoji or emoji ID found")

        if ctx.cache and (emoji := ctx.cache.get_emoji(emoji_id)):
            return emoji

        if cache:
            try:
                return await cache.get(emoji_id)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find emoji") from None

            except async_cache.CacheMissError:
                pass

        if ctx.guild_id:
            try:
                return await ctx.rest.fetch_emoji(ctx.guild_id, emoji_id)

            except hikari.NotFoundError:
                pass

        raise ValueError("Couldn't find emoji")


@typing_extensions.deprecated("Use ToEmoji")
class EmojiConverter(ToEmoji):
    """Deprecated alias of [tanjun.conversion.ToEmoji][]."""

    __slots__ = ()


_GuildCacheT = async_cache.SfCache[hikari.Guild]


class ToGuild(BaseConverter):
    """Standard converter for guilds.

    For a standard instance of this see [tanjun.conversion.to_guild][].
    """

    __slots__ = ()

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        # <<inherited docstring from BaseConverter>>.
        return [(_GuildCacheT, hikari.api.CacheComponents.GUILDS, hikari.Intents.GUILDS)]

    async def __call__(
        self,
        argument: _SnowflakeIsh,
        /,
        ctx: alluka.Injected[tanjun.Context],
        *,
        cache: alluka.Injected[typing.Optional[_GuildCacheT]] = None,
    ) -> hikari.Guild:
        guild_id = parse_snowflake(argument, message="No valid guild ID found")
        if ctx.cache and (guild := ctx.cache.get_guild(guild_id)):
            return guild

        if cache:
            try:
                return await cache.get(guild_id)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find guild") from None

            except async_cache.CacheMissError:
                pass

        try:
            return await ctx.rest.fetch_guild(guild_id)

        except hikari.NotFoundError:
            pass

        raise ValueError("Couldn't find guild")


@typing_extensions.deprecated("Use ToGuild")
class GuildConverter(ToGuild):
    """Deprecated alias of [tanjun.conversion.ToGuild][]."""

    __slots__ = ()


_InviteCacheT = async_cache.AsyncCache[str, hikari.InviteWithMetadata]


class ToInvite(BaseConverter):
    """Standard converter for invites."""

    __slots__ = ()

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        # <<inherited docstring from BaseConverter>>.
        return [(_InviteCacheT, hikari.api.CacheComponents.INVITES, hikari.Intents.GUILD_INVITES)]

    async def __call__(
        self,
        argument: str,
        /,
        ctx: alluka.Injected[tanjun.Context],
        *,
        cache: alluka.Injected[typing.Optional[_InviteCacheT]] = None,
    ) -> hikari.Invite:
        if ctx.cache and (invite := ctx.cache.get_invite(argument)):
            return invite

        if cache:
            try:
                return await cache.get(argument)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find invite") from None

            except async_cache.CacheMissError:
                pass

        try:
            return await ctx.rest.fetch_invite(argument)
        except hikari.NotFoundError:
            pass

        raise ValueError("Couldn't find invite")


@typing_extensions.deprecated("Use ToInvite")
class InviteConverter(ToInvite):
    """Deprecated alias of [tanjun.conversion.ToInvite][]."""

    __slots__ = ()


class ToInviteWithMetadata(BaseConverter):
    """Standard converter for invites with metadata.

    For a standard instance of this see [tanjun.conversion.to_invite_with_metadata][].

    !!! note
        Unlike [tanjun.conversion.InviteConverter][], this converter is cache dependent.
    """

    __slots__ = ()

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        # <<inherited docstring from BaseConverter>>.
        return [(_InviteCacheT, hikari.api.CacheComponents.INVITES, hikari.Intents.GUILD_INVITES)]

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return True

    async def __call__(
        self,
        argument: str,
        /,
        ctx: alluka.Injected[tanjun.Context],
        *,
        cache: alluka.Injected[typing.Optional[_InviteCacheT]] = None,
    ) -> hikari.InviteWithMetadata:
        if ctx.cache and (invite := ctx.cache.get_invite(argument)):
            return invite

        if cache and (invite := await cache.get(argument)):
            return invite

        raise ValueError("Couldn't find invite")


@typing_extensions.deprecated("Use ToInviteWithMetadata")
class InviteWithMetadataConverter(ToInviteWithMetadata):
    """Deprecated alias of [tanjun.conversion.ToInviteWithMetadata][]."""

    __slots__ = ()


_MemberCacheT = async_cache.SfGuildBound[hikari.Member]


class ToMember(BaseConverter):
    """Standard converter for guild members.

    For a standard instance of this see [tanjun.conversion.to_member][].

    This converter allows both mentions, raw IDs and partial usernames/nicknames
    and only works within a guild context.
    """

    __slots__ = ()

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        # <<inherited docstring from BaseConverter>>.
        return [
            (_MemberCacheT, hikari.api.CacheComponents.MEMBERS, hikari.Intents.GUILD_MEMBERS | hikari.Intents.GUILDS)
        ]

    async def __call__(
        self,
        argument: _SnowflakeIsh,
        /,
        ctx: alluka.Injected[tanjun.Context],
        *,
        cache: alluka.Injected[typing.Optional[_MemberCacheT]] = None,
    ) -> hikari.Member:
        if ctx.guild_id is None:
            raise ValueError("Cannot get a member from a DM channel")

        try:
            user_id = parse_user_id(argument, message="No valid user mention or ID found")

        except ValueError:
            if isinstance(argument, str):
                try:
                    return (await ctx.rest.search_members(ctx.guild_id, argument))[0]

                except (hikari.NotFoundError, IndexError):
                    pass

        else:
            if ctx.cache and (member := ctx.cache.get_member(ctx.guild_id, user_id)):
                return member

            if cache:
                try:
                    return await cache.get_from_guild(ctx.guild_id, user_id)

                except async_cache.EntryNotFound:
                    raise ValueError("Couldn't find member in this guild") from None

                except async_cache.CacheMissError:
                    pass

            try:
                return await ctx.rest.fetch_member(ctx.guild_id, user_id)

            except hikari.NotFoundError:
                pass

        raise ValueError("Couldn't find member in this guild")


@typing_extensions.deprecated("Use ToMember")
class MemberConverter(ToMember):
    """Deprecated alias of [tanjun.conversion.ToMember][]."""

    __slots__ = ()


_PresenceCacheT = async_cache.SfGuildBound[hikari.MemberPresence]


class ToPresence(BaseConverter):
    """Standard converter for presences.

    For a standard instance of this see [tanjun.conversion.to_presence][].

    This converter is cache dependent and only works in a guild context.
    """

    __slots__ = ()

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        # <<inherited docstring from BaseConverter>>.
        return [
            (
                _PresenceCacheT,
                hikari.api.CacheComponents.PRESENCES,
                hikari.Intents.GUILD_PRESENCES | hikari.Intents.GUILDS,
            )
        ]

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return True

    async def __call__(
        self,
        argument: _SnowflakeIsh,
        /,
        ctx: alluka.Injected[tanjun.Context],
        *,
        cache: alluka.Injected[typing.Optional[_PresenceCacheT]] = None,
    ) -> hikari.MemberPresence:
        if ctx.guild_id is None:
            raise ValueError("Cannot get a presence from a DM channel")

        user_id = parse_user_id(argument, message="No valid member mention or ID found")
        if ctx.cache and (presence := ctx.cache.get_presence(ctx.guild_id, user_id)):
            return presence

        if cache and (presence := await cache.get_from_guild(ctx.guild_id, user_id, default=None)):
            return presence

        raise ValueError("Couldn't find presence in current guild")


@typing_extensions.deprecated("Use ToPresence")
class PresenceConverter(ToPresence):
    """Deprecated alias of [tanjun.conversion.ToPresence][]."""

    __slots__ = ()


_RoleCacheT = async_cache.SfCache[hikari.Role]


class ToRole(BaseConverter):
    """Standard converter for guild roles.

    For a standard instance of this see [tanjun.conversion.to_role][].
    """

    __slots__ = ()

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        return [(_RoleCacheT, hikari.api.CacheComponents.ROLES, hikari.Intents.GUILDS)]

    async def __call__(
        self,
        argument: _SnowflakeIsh,
        /,
        ctx: alluka.Injected[tanjun.Context],
        *,
        cache: alluka.Injected[typing.Optional[_RoleCacheT]] = None,
    ) -> hikari.Role:
        role_id = parse_role_id(argument, message="No valid role mention or ID found")
        if ctx.cache and (role := ctx.cache.get_role(role_id)):
            return role

        if cache:
            try:
                return await cache.get(role_id)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find role") from None

            except async_cache.CacheMissError:
                pass

        if ctx.guild_id:
            for role in await ctx.rest.fetch_roles(ctx.guild_id):
                if role.id == role_id:
                    return role

        raise ValueError("Couldn't find role")


@typing_extensions.deprecated("Use ToRole")
class RoleConverter(ToRole):
    """Deprecated alias of [tanjun.conversion.ToRole][]."""

    __slots__ = ()


_UserCacheT = async_cache.SfCache[hikari.User]


class ToUser(BaseConverter):
    """Standard converter for users.

    For a standard instance of this see [tanjun.conversion.to_user][].
    """

    __slots__ = ()

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        # <<inherited docstring from BaseConverter>>.
        return [(_UserCacheT, hikari.api.CacheComponents.NONE, hikari.Intents.GUILDS | hikari.Intents.GUILD_MEMBERS)]

    async def __call__(
        self,
        argument: _SnowflakeIsh,
        /,
        ctx: alluka.Injected[tanjun.Context],
        *,
        cache: alluka.Injected[typing.Optional[_UserCacheT]] = None,
    ) -> hikari.User:
        # TODO: search by name if this is a guild context
        user_id = parse_user_id(argument, message="No valid user mention or ID found")
        if ctx.cache and (user := ctx.cache.get_user(user_id)):
            return user

        if cache:
            try:
                return await cache.get(user_id)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find user") from None

            except async_cache.CacheMissError:
                pass

        try:
            return await ctx.rest.fetch_user(user_id)

        except hikari.NotFoundError:
            pass

        raise ValueError("Couldn't find user")


@typing_extensions.deprecated("Use ToUser")
class UserConverter(ToUser):
    """Deprecated alias of [tanjun.conversion.ToUser][]."""

    __slots__ = ()


_MessageCacheT = async_cache.SfCache[hikari.Message]


class ToMessage(BaseConverter):
    """Standard converter for messages.

    For a standard instance of this see [tanjun.conversion.to_message][].
    """

    __slots__ = ()

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        # <<inherited docstring from BaseConverter>>.
        return [
            (
                _MessageCacheT,
                hikari.api.CacheComponents.MESSAGES,
                hikari.Intents.GUILD_MESSAGES | hikari.Intents.DM_MESSAGES,
            )
        ]

    async def __call__(
        self,
        argument: _SnowflakeIsh,
        /,
        ctx: alluka.Injected[tanjun.Context],
        *,
        cache: alluka.Injected[typing.Optional[_MessageCacheT]] = None,
    ) -> hikari.Message:
        channel_id, message_id = parse_message_id(argument)
        if ctx.cache and (message := ctx.cache.get_message(message_id)):
            return message

        if cache:
            try:
                return await cache.get(message_id)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find message") from None

            except async_cache.CacheMissError:
                pass

        try:
            return await ctx.rest.fetch_message(channel_id or ctx.channel_id, message_id)

        except hikari.NotFoundError:
            pass

        raise ValueError("Couldn't find message")


_VoiceStateCacheT = async_cache.SfGuildBound[hikari.VoiceState]


class ToVoiceState(BaseConverter):
    """Standard converter for voice states.

    For a standard instance of this see [tanjun.conversion.to_voice_state][].

    !!! note
        This converter is cache dependent and only works in a guild context.
    """

    __slots__ = ()

    @property
    def caches(self) -> collections.Sequence[tuple[typing.Any, hikari.api.CacheComponents, hikari.Intents]]:
        # <<inherited docstring from BaseConverter>>.
        return [
            (
                _VoiceStateCacheT,
                hikari.api.CacheComponents.VOICE_STATES,
                hikari.Intents.GUILD_VOICE_STATES | hikari.Intents.GUILDS,
            )
        ]

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return True

    async def __call__(
        self,
        argument: _SnowflakeIsh,
        /,
        ctx: alluka.Injected[tanjun.Context],
        *,
        cache: alluka.Injected[typing.Optional[_VoiceStateCacheT]] = None,
    ) -> hikari.VoiceState:
        if ctx.guild_id is None:
            raise ValueError("Cannot get a voice state from a DM channel")

        user_id = parse_user_id(argument, message="No valid user mention or ID found")

        if ctx.cache and (state := ctx.cache.get_voice_state(ctx.guild_id, user_id)):
            return state

        if cache and (state := await cache.get_from_guild(ctx.guild_id, user_id, default=None)):
            return state

        raise ValueError("Voice state couldn't be found for current guild")


@typing_extensions.deprecated("Use ToVoiceState")
class VoiceStateConverter(ToVoiceState):
    """Deprecated alias of [tanjun.conversion.ToVoiceState][]."""

    __slots__ = ()


class _IDMatcherSigProto(typing.Protocol):
    def __call__(self, value: _SnowflakeIsh, /, *, message: str = "No valid mention or ID found") -> hikari.Snowflake:
        raise NotImplementedError


def _make_snowflake_parser(regex: re.Pattern[str], /) -> _IDMatcherSigProto:
    def parse(value: _SnowflakeIsh, /, *, message: str = "No valid mention or ID found") -> hikari.Snowflake:
        """Parse a snowflake from a string or int value.

        !!! note
            This only allows the relevant entity's mention format if applicable.

        Parameters
        ----------
        value
            The value to parse (this argument can only be passed positionally).
        message
            The error message to raise if the value cannot be parsed.

        Returns
        -------
        hikari.Snowflake
            The parsed snowflake.

        Raises
        ------
        ValueError
            If the value cannot be parsed.
        """
        result: typing.Optional[hikari.Snowflake] = None
        if isinstance(value, int) or value.isdigit():
            result = hikari.Snowflake(value)

        else:
            capture = next(regex.finditer(value), None)
            result = hikari.Snowflake(capture.groups()[0]) if capture else None

        # We should also range check the provided ID.
        if result is not None and _range_check(result):
            return result

        raise ValueError(message) from None

    return parse


_IDSearcherSig = collections.Callable[[_SnowflakeIsh], list[hikari.Snowflake]]


def _range_check(snowflake: hikari.Snowflake, /) -> bool:
    return snowflake.min() <= snowflake <= snowflake.max()


def _make_snowflake_searcher(regex: re.Pattern[str], /) -> _IDSearcherSig:
    def parse(value: _SnowflakeIsh, /) -> list[hikari.Snowflake]:
        """Get the snowflakes in a string.

        !!! note
            This only allows the relevant entity's mention format if applicable.

        Parameters
        ----------
        value
            The value to parse (this argument can only be passed positionally).

        Returns
        -------
        list[hikari.Snowflake]
            List of the IDs found in the string.
        """
        if isinstance(value, str):
            if value.isdigit() and _range_check(result := hikari.Snowflake(value)):
                return [result]

            results = filter(
                _range_check, map(hikari.Snowflake, (match.groups()[0] for match in regex.finditer(value)))
            )
            return [*results, *filter(_range_check, map(hikari.Snowflake, filter(str.isdigit, value.split())))]

        try:
            result = hikari.Snowflake(value)

        except ValueError:
            return []

        if _range_check(result):
            return [result]

        return []

    return parse


_SNOWFLAKE_REGEX = re.compile(r"<[@&?!#a]{0,3}(?::\w+:)?(\d+)>")
parse_snowflake: _IDMatcherSigProto = _make_snowflake_parser(_SNOWFLAKE_REGEX)
"""Parse a snowflake from a string or int value.

Parameters
----------
value : str | int | float
    The value to parse (this argument can only be passed positionally).
message : str
    The error message to raise if the value cannot be parsed.

Returns
-------
hikari.Snowflake
    The parsed snowflake.

Raises
------
ValueError
    If the value cannot be parsed.
"""

search_snowflakes: _IDSearcherSig = _make_snowflake_searcher(_SNOWFLAKE_REGEX)
"""Get the snowflakes in a string.

Parameters
----------
value : str | int | float
    The value to parse (this argument can only be passed positionally).

Returns
-------
list[hikari.Snowflake]
    List of the snowflakes in the string.
"""

_CHANNEL_ID_REGEX = re.compile(r"<#(\d+)>")
parse_channel_id: _IDMatcherSigProto = _make_snowflake_parser(_CHANNEL_ID_REGEX)
"""Parse a channel ID from a string or int value.

Parameters
----------
value : str | int | float
    The value to parse (this argument can only be passed positionally).
message : str
    The error message to raise if the value cannot be parsed.

Returns
-------
hikari.Snowflake
    The parsed channel ID.

Raises
------
ValueError
    If the value cannot be parsed.
"""

search_channel_ids: _IDSearcherSig = _make_snowflake_searcher(_CHANNEL_ID_REGEX)
"""Get the channel IDs in a string.

Parameters
----------
value : str | int | float
    The value to parse (this argument can only be passed positionally).

Returns
-------
list[hikari.Snowflake]
    List of the channel IDs in the string.
"""

_EMOJI_ID_REGEX = re.compile(r"<a?:\w+:(\d+)>")
parse_emoji_id: _IDMatcherSigProto = _make_snowflake_parser(_EMOJI_ID_REGEX)
"""Parse an Emoji ID from a string or int value.

Parameters
----------
value : str | int | float
    The value to parse (this argument can only be passed positionally).
message : str
    The error message to raise if the value cannot be parsed.

    Defaults to "No valid mention or ID found".

Returns
-------
hikari.Snowflake
    The parsed Emoji ID.

Raises
------
ValueError
    If the value cannot be parsed.
"""

search_emoji_ids: _IDSearcherSig = _make_snowflake_searcher(_EMOJI_ID_REGEX)
"""Get the emoji IDs in a string.

Parameters
----------
value : str | int | float
    The value to parse (this argument can only be passed positionally).

Returns
-------
list[hikari.Snowflake]
    List of the emoji IDs in the string.
"""

_ROLE_ID_REGEX = re.compile(r"<@&(\d+)>")
parse_role_id: _IDMatcherSigProto = _make_snowflake_parser(_ROLE_ID_REGEX)
"""Parse a role ID from a string or int value.

Parameters
----------
value : str | int | float
    The value to parse (this argument can only be passed positionally).
message : str
    The error message to raise if the value cannot be parsed.

    Defaults to "No valid mention or ID found".

Returns
-------
hikari.Snowflake
    The parsed role ID.

Raises
------
ValueError
    If the value cannot be parsed.
"""

search_role_ids: _IDSearcherSig = _make_snowflake_searcher(_ROLE_ID_REGEX)
"""Get the role IDs in a string.

Parameters
----------
value : str | int | float
    The value to parse (this argument can only be passed positionally).

Returns
-------
list[hikari.Snowflake]
    List of the role IDs in the string.
"""

_USER_ID_REGEX = re.compile(r"<@!?(\d+)>")
parse_user_id: _IDMatcherSigProto = _make_snowflake_parser(_USER_ID_REGEX)
"""Parse a user ID from a string or int value.

Parameters
----------
value : str | int | float
    The value to parse (this argument can only be passed positionally).
message : str
    The error message to raise if the value cannot be parsed.

    Defaults to "No valid mention or ID found".

Returns
-------
hikari.Snowflake
    The parsed user ID.

Raises
------
ValueError
    If the value cannot be parsed.
"""

search_user_ids: _IDSearcherSig = _make_snowflake_searcher(_USER_ID_REGEX)
"""Get the user IDs in a string.

Parameters
----------
value : str | int | float
    The value to parse (this argument can only be passed positionally).

Returns
-------
list[hikari.Snowflake]
    List of the user IDs in the string.
"""

_MESSAGE_LINK_REGEX = re.compile(r"(\d+|@me)/(\d+)/(\d+)")


def parse_message_id(
    value: _SnowflakeIsh, /, *, message: str = "No valid message link or ID found"
) -> tuple[typing.Optional[hikari.Snowflake], hikari.Snowflake]:
    """Parse a user ID from a string or int value.

    Parameters
    ----------
    value
        The value to parse (this argument can only be passed positionally).
    message
        The error message to raise if the value cannot be parsed.

    Returns
    -------
    tuple[hikari.Snowflake | None, hikari.Snowflake]
        The parsed channel and message IDs.

    Raises
    ------
    ValueError
        If the value cannot be parsed.
    """
    channel_id: typing.Optional[hikari.Snowflake] = None
    message_id: typing.Optional[hikari.Snowflake] = None

    if isinstance(value, int) or value.isdigit():
        message_id = hikari.Snowflake(value)

    else:
        capture = next(_MESSAGE_LINK_REGEX.finditer(value), None)
        if capture:
            channel_id = hikari.Snowflake(capture[2])
            message_id = hikari.Snowflake(capture[3])

    # We should also range check the provided ID.
    if channel_id is not None and not _range_check(channel_id):
        raise ValueError(message) from None

    if message_id is not None and _range_check(message_id):
        return channel_id, message_id

    raise ValueError(message) from None


def _build_url_parser(callback: collections.Callable[[str], _ValueT], /) -> collections.Callable[[str], _ValueT]:
    def parse(value: str, /) -> _ValueT:
        """Convert an argument to a [urllib.parse][] type.

        Parameters
        ----------
        value
            The value to parse (this argument can only be passed positionally).

        Returns
        -------
        _ValueT
            The parsed URL.

        Raises
        ------
        ValueError
            If the argument couldn't be parsed.
        """
        if value.startswith("<") and value.endswith(">"):
            value = value[1:-1]

        return callback(value)

    return parse


defragment_url: collections.Callable[[str], urlparse.DefragResult] = _build_url_parser(urlparse.urldefrag)
"""Convert an argument to a defragmented URL.

Parameters
----------
value : str
    The value to parse (this argument can only be passed positionally).

Returns
-------
urllib.parse.DefragResult
    The parsed URL.

Raises
------
ValueError
    If the argument couldn't be parsed.
"""

parse_url: collections.Callable[[str], urlparse.ParseResult] = _build_url_parser(urlparse.urlparse)
"""Convert an argument to a parsed URL.

Parameters
----------
value : str
    The value to parse (this argument can only be passed positionally).

Returns
-------
urllib.parse.ParseResult
    The parsed URL.

Raises
------
ValueError
    If the argument couldn't be parsed.
"""


split_url: collections.Callable[[str], urlparse.SplitResult] = _build_url_parser(urlparse.urlsplit)
"""Convert an argument to a split URL.

Parameters
----------
value : str
    The value to parse (this argument can only be passed positionally).

Returns
-------
urllib.parse.SplitResult
    The split URL.

Raises
------
ValueError
    If the argument couldn't be parsed.
"""

_DATETIME_REGEX = re.compile(r"<-?t:(\d+)(?::\w)?>")


def to_datetime(value: str, /) -> datetime.datetime:
    """Parse a datetime from Discord's datetime format.

    More information on this format can be found at
    <https://discord.com/developers/docs/reference#message-formatting-timestamp-styles>

    Parameters
    ----------
    value
        The value to parse.

    Returns
    -------
    datetime.datetime
        The parsed datetime.

    Raises
    ------
    ValueError
        If the value cannot be parsed.
    """
    try:
        timestamp = int(next(_DATETIME_REGEX.finditer(value)).groups()[0])

    except StopIteration:
        raise ValueError("Not a valid datetime") from None

    return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)


_VALID_DATETIME_STYLES = frozenset(("t", "T", "d", "D", "f", "F", "R"))


@typing.overload
def from_datetime(value: datetime.timedelta, /) -> str:
    ...


@typing.overload
def from_datetime(value: datetime.datetime, /, *, style: str = "f") -> str:
    ...


def from_datetime(value: typing.Union[datetime.datetime, datetime.timedelta], /, *, style: str = "f") -> str:
    """Format a datetime as Discord's datetime format.

    More information on this format can be found at
    <https://discord.com/developers/docs/reference#message-formatting-timestamp-styles>

    Parameters
    ----------
    value
        The datetime to format.

        If a timedelta is passed here then this is treated as a date that's
        relative to the current time.
    style
        The style to use.

        The valid styles can be found at
        <https://discord.com/developers/docs/reference#message-formatting-formats>.

        This is always "R" when `value` is a [datetime.timedelta][].

    Returns
    -------
    str
        The formatted datetime.

    Raises
    ------
    ValueError
        If the provided datetime is timezone naive.
        If an invalid style is provided.
    """
    if isinstance(value, datetime.timedelta):
        return from_datetime(datetime.datetime.now(tz=datetime.timezone.utc) + value, style="R")

    if style not in _VALID_DATETIME_STYLES:
        raise ValueError(f"Invalid style: {style}")

    if value.tzinfo is None:
        raise ValueError("Cannot convert naive datetimes, please specify a timezone.")

    return f"<t:{round(value.timestamp())}:{style}>"


_YES_VALUES = frozenset(("y", "yes", "t", "true", "on", "1"))
_NO_VALUES = frozenset(("n", "no", "f", "false", "off", "0"))


def to_bool(value: str, /) -> bool:
    """Convert user string input into a boolean value.

    Parameters
    ----------
    value
        The value to convert.

    Returns
    -------
    bool
        The converted value.

    Raises
    ------
    ValueError
        If the value cannot be converted.
    """
    value = value.lower().strip()
    if value in _YES_VALUES:
        return True

    if value in _NO_VALUES:
        return False

    raise ValueError(f"Invalid bool value `{value}`")


def to_color(argument: _SnowflakeIsh, /) -> hikari.Color:
    """Convert user input to a [hikari.colors.Color][] object."""
    if isinstance(argument, str):
        values = argument.split(" ")
        if all(value.isdigit() for value in values):
            if len(values) == 1:
                return hikari.Color.of(int(values[0]))

            return hikari.Color.of(list(map(int, values)))

        if len(values) == 1:
            return hikari.Color.of(values[0])

        raise ValueError("Not a valid color representation")  # noqa: TC004

    return hikari.Color.of(argument)


_TYPE_OVERRIDES: dict[collections.Callable[..., typing.Any], collections.Callable[[str], typing.Any]] = {
    bool: to_bool,
    bytes: lambda d: bytes(d, "utf-8"),
    bytearray: lambda d: bytearray(d, "utf-8"),
    datetime.datetime: to_datetime,
    hikari.Snowflake: parse_snowflake,
    urlparse.DefragResult: defragment_url,
    urlparse.ParseResult: parse_url,
    urlparse.SplitResult: split_url,
}


def override_type(cls: parsing.ConverterSig[typing.Any], /) -> parsing.ConverterSig[typing.Any]:  # noqa: D103
    return _TYPE_OVERRIDES.get(cls, cls)


to_channel: typing.Final[ToChannel] = ToChannel()
"""Convert user input to a [hikari.channels.PartialChannel][] object."""

to_colour: typing.Final[collections.Callable[[_SnowflakeIsh], hikari.Color]] = to_color
"""Convert user input to a [hikari.colors.Color][] object."""

to_emoji: typing.Final[ToEmoji] = ToEmoji()
"""Convert user input to a cached [hikari.emojis.KnownCustomEmoji][] object.

!!! note
    If you just want to convert input to a [hikari.emojis.Emoji][],
    [hikari.emojis.CustomEmoji][] or [hikari.emojis.UnicodeEmoji][] without
    making any cache or REST calls then you can just use the relevant
    [hikari.emojis.Emoji.parse][], [hikari.emojis.CustomEmoji.parse][] or
    [hikari.emojis.UnicodeEmoji.parse][] methods.
"""

to_guild: typing.Final[ToGuild] = ToGuild()
"""Convert user input to a [hikari.guilds.Guild][] object."""

to_invite: typing.Final[ToInvite] = ToInvite()
"""Convert user input to a cached [hikari.invites.InviteWithMetadata][] object."""

to_invite_with_metadata: typing.Final[ToInviteWithMetadata] = ToInviteWithMetadata()
"""Convert user input to a [hikari.invites.Invite][] object."""

to_member: typing.Final[ToMember] = ToMember()
"""Convert user input to a [hikari.guilds.Member][] object."""

to_presence: typing.Final[ToPresence] = ToPresence()
"""Convert user input to a cached [hikari.presences.MemberPresence][]."""

to_role: typing.Final[ToRole] = ToRole()
"""Convert user input to a [hikari.guilds.Role][] object."""

to_snowflake: typing.Final[collections.Callable[[_SnowflakeIsh], hikari.Snowflake]] = parse_snowflake
"""Convert user input to a [hikari.snowflakes.Snowflake][].

!!! note
    This also range validates the input.
"""

to_user: typing.Final[ToUser] = ToUser()
"""Convert user input to a [hikari.users.User][] object."""

to_message: typing.Final[ToMessage] = ToMessage()
"""Convert user input to a [hikari.messages.Message][] object."""

to_voice_state: typing.Final[ToVoiceState] = ToVoiceState()
"""Convert user input to a cached [hikari.voices.VoiceState][]."""
